__version__ = "1.5.0"

from . import basics, model
from .bins import *
from .example_data import *
from .notebook import *
from .props import *
from .stats import *
from .utils import *


# Create wrapper functions with deprecation errors
def _make_wrapper(func_name, module):
    def wrapper(*args, **kwargs):
        raise DeprecationWarning(
            f"{func_name}() is deprecated. Use {module}.{func_name}() instead."
        )
    return wrapper

# Model functions that need wrappers
model_functions = [
    "distr_plot",
    "make_train",
    "cross_validation",
    "regress",
    "rforest",
    "logistic",
    "mlp",
    "xgboost",
    "gains_plot"
]

# Basics classes that need wrappers
basics_functions = [
    "central_limit_theorem",
    "compare_means",
    "compare_props",
    "correlation",
    "cross_tabs",
    "goodness",
    "prob_calc",
    "single_mean",
    "single_prop"
]

# Create wrappers for model functions
for func_name in model_functions:
    globals()[func_name] = _make_wrapper(func_name, "model")

# Create wrappers for basics classes
for func_name in basics_functions:
    globals()[func_name] = _make_wrapper(func_name, "basics")

__all__ = ['model', 'basics'] + model_functions + basics_functions



from pathlib import Path

from shiny import App
from starlette.applications import Starlette
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles

import pyrsm.radiant.utils as ru
from pyrsm.radiant.model.regress import model_regress

www_dir = Path(__file__).parent.parent.parent / "radiant" / "www"
app_static = StaticFiles(directory=www_dir, html=False)

# data_dct, descriptions_dct = ru.get_dfs(pkg="model")
data_dct, descriptions_dct = ru.get_dfs(pkg="model", name="diamonds")
data_dct.update({"diamonds100": data_dct["diamonds"].sample(100)})
descriptions_dct.update({"diamonds100": descriptions_dct["diamonds"]})
rc = model_regress(data_dct, descriptions_dct, state=None, code=True)

routes = [
    Mount("/www", app=app_static),
    Mount("/", app=App(rc.shiny_ui, rc.shiny_server, debug=False)),
]
app = Starlette(debug=True, routes=routes)



from cmath import sqrt
from typing import Optional, Union

import matplotlib.pyplot as plt
import pandas as pd
import polars as pl
from scipy.stats import chisquare

from pyrsm.radiant import utils as ru
from pyrsm.utils import check_dataframe, ifelse


class goodness:
    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        var: str,
        probs: Optional[tuple[float, ...]] = None,
        figsize: tuple[float, float] = None,
    ) -> None:
        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name]
        else:
            self.data = data
            self.name = "Not provided"

        self.data = check_dataframe(self.data)
        self.var = var
        self.figsize = figsize
        self.probs = probs

        self.freq = self.data[self.var].value_counts().to_dict()
        self.nlev = len(self.freq)
        if self.probs is None:
            self.probs = [1 / self.nlev] * self.nlev

        self.observed = pd.DataFrame(
            {k: [v] for k, v in self.freq.items()},
            columns=sorted(self.freq.keys()),
        )
        self.observed["Total"] = self.observed[list(self.observed.columns)].sum(axis=1)
        self.expected = pd.DataFrame(
            {
                sorted(self.freq.keys())[i]: [
                    self.probs[i] * self.observed.at[0, "Total"],
                ]
                for i in range(len(self.freq.keys()))
            },
            columns=sorted(self.freq.keys()),
        )
        self.expected["Total"] = self.expected[list(self.expected.columns)].sum(axis=1)
        self.chisq = pd.DataFrame(
            {
                column: [
                    round(
                        (
                            (self.observed.at[0, column] - self.expected.at[0, column])
                            ** 2
                        )
                        / self.expected.at[0, column],
                        2,
                    ),
                ]
                for column in self.expected.columns.tolist()[:-1]
            },
            columns=self.expected.columns.tolist(),
        )
        self.chisq["Total"] = self.chisq[list(self.chisq.columns)].sum(axis=1)

        self.stdev = pd.DataFrame(
            {
                column: [
                    round(
                        (self.observed.at[0, column] - self.expected.at[0, column])
                        / sqrt(self.expected.at[0, column]).real,
                        2,
                    ),
                ]
                for column in self.expected.columns.tolist()[:-1]
            },
            columns=self.expected.columns.tolist()[:-1],
        )

    def summary(
        self, output: list[str] = ["observed", "expected"], dec: int = 3
    ) -> None:
        pd.set_option("display.max_columns", 20)
        pd.set_option("display.max_rows", 20)

        output = ifelse(isinstance(output, str), [output], output)

        print("Goodness of fit test")
        print(f"Data         : {self.name}")
        if self.var not in self.data.columns:
            raise ValueError(f"{self.var} does not exist in chosen dataset")

        print(f"Variable     : {self.var}")
        if self.nlev != len(self.probs):
            raise ValueError(
                f'Number of elements in "probs" should match the number of levels in {self.var} ({self.nlev})'
            )

        if not 0.999 <= sum(self.probs) <= 1.001:
            raise ValueError("Probabilities do not sum to 1 ({sum(self.probs)})")

        print(f'Probabilities: {" ".join(map(str, self.probs))}')
        print(
            f"Null hyp.    : The distribution of {self.var} is consistent with the specified distribution"
        )
        print(
            f"Alt. hyp.    : The distribution of {self.var} is not consistent with the specified distribution"
        )

        if "observed" in output:
            print("\nObserved:")
            print(self.observed.to_string(index=False))

        if "expected" in output:
            print("\nExpected: total x p")
            print(self.expected.to_string(index=False))

        if "chisq" in output:
            print("\nContribution to chi-squared: (observed - expected) ^ 2 / expected")
            print(self.chisq.to_string(index=False))

        if "dev_std" in output:
            print("\nDeviation standardized: (observed - expected) / sqrt(expected)\n")
            print(self.stdev.to_string(index=False))

        chisq, p_val = chisquare(
            [self.freq[key] for key in sorted(self.freq.keys())],
            [self.expected.at[0, key] for key in sorted(self.freq.keys())],
        )

        if p_val < 0.001:
            p_val = "< .001"
        else:
            p_val = round(p_val, dec)
        print(
            f"\nChi-squared: {round(chisq, dec)} df ({self.nlev - 1}), p.value {p_val}"
        )

    def plot(self, plots: list[str] = [], **kwargs) -> None:
        plots = ifelse(isinstance(plots, str), [plots], plots)

        args = {"rot": False}
        if "observed" in plots:
            tab = self.observed.copy().drop(columns="Total").transpose()
            args["title"] = "Observed frequencies"
            args.update(**kwargs)
            fig = tab.plot.bar(**args, legend=False)
        if "expected" in plots:
            tab = self.expected.copy().drop(columns="Total").transpose()
            args["title"] = "Expected frequencies"
            args.update(**kwargs)
            fig = tab.plot.bar(**args, legend=False)
        if "chisq" in plots:
            tab = self.chisq.drop(columns="Total").transpose()
            args["title"] = "Contribution to chi-squared statistic"
            args.update(**kwargs)
            fig = tab.plot.bar(**args, legend=False)
        if "dev_std" in plots:
            tab = self.stdev.transpose()
            args["title"] = "Deviation standardized"
            args.update(**kwargs, legend=False)
            fig, ax = plt.subplots()
            tab.plot.bar(**args, ax=ax)
            ax.axhline(y=1.96, color="black", linestyle="--")
            ax.axhline(y=1.64, color="black", linestyle="--")
            ax.axhline(y=0, color="black")
            ax.axhline(y=-1.96, color="black", linestyle="--")
            ax.axhline(y=-1.64, color="black", linestyle="--")
            ax.annotate("95%", xy=(0, 2.1), va="bottom", ha="center")
            ax.annotate("90%", xy=(0, 1.4), va="top", ha="center")


if __name__ == "__main__":
    import pyrsm as rsm

    data_dct, descriptions_dct = ru.get_dfs(pkg="basics", name="newspaper")
    gf = rsm.basics.goodness(data=data_dct, var="Income", probs=[1 / 2, 1 / 2])
    gf.summary(output=["observed"])
    gf.plot(plots=["observed", "expected", "chisq", "dev_std"])



from cmath import sqrt
from typing import Union

import numpy as np
import pandas as pd
import polars as pl
from scipy import stats

import pyrsm.basics.utils as bu
from pyrsm.model.model import sig_stars
from pyrsm.utils import check_dataframe, ifelse


class single_prop:
    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        var: str,
        lev: str = None,
        alt_hyp: str = "two-sided",
        conf: float = 0.95,
        comp_value: float = 0.5,
        test_type: str = "binomial",
    ) -> None:
        if comp_value == 0 or comp_value == 1:
            raise Exception("Please choose a comparison value between 0 and 1")

        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name].copy()
        else:
            self.data = data
            self.name = "Not provided"
        self.data = check_dataframe(self.data)
        self.var = var
        self.lev = lev
        self.alt_hyp = alt_hyp
        self.conf = conf
        self.alpha = 1 - self.conf
        self.comp_value = comp_value
        self.test_type = test_type
        self.ns = len(self.data[self.data[self.var] == self.lev])
        self.n_missing = self.data[self.var].isna().sum()
        self.n = len(self.data) - self.n_missing
        self.p = self.ns / self.n
        self.sd = sqrt(self.p * (1 - self.p)).real
        self.se = (self.sd / sqrt(self.n)).real
        self.se_p0 = sqrt(self.comp_value * (1 - self.comp_value) / self.n).real
        self.z_critical = stats.norm.ppf(1 - self.alpha / 2)
        self.z_score = None
        self.me = self.z_critical * self.se
        self.diff = self.p - self.comp_value

        def wilson_ci(zc):
            # Wilson CI used by R's prop.test function
            z_square_n = zc**2 / self.n
            denominator = 1 + z_square_n
            lower_limit = (
                self.p
                + z_square_n / 2
                - zc * np.sqrt(self.se**2 + z_square_n / (4 * self.n))
            ) / denominator
            upper_limit = (
                self.p
                + z_square_n / 2
                + zc * np.sqrt(self.se**2 + z_square_n / (4 * self.n))
            ) / denominator
            return [lower_limit, upper_limit]

        if test_type == "binomial":
            result = stats.binomtest(self.ns, self.n, self.comp_value, self.alt_hyp)
            self.ci = result.proportion_ci(confidence_level=conf)
            self.p_val = result.pvalue
        else:
            self.z_score = (self.p - self.comp_value) / self.se_p0
            # self.z_score = (self.p - self.comp_value) / self.se_p
            p_val = stats.norm.cdf(self.z_score)
            if self.alt_hyp == "two-sided":
                self.p_val = p_val * 2
                # traditional CI calculation
                # me = self.z_critical * self.se
                # self.ci = [self.p - me, self.p + me]
                self.ci = wilson_ci(self.z_critical)
            elif self.alt_hyp == "less":
                self.p_val = p_val
                # traditional CI calculation
                # z_critical = stats.norm.ppf(self.conf)
                # me = z_critical * self.se
                # self.ci = [0, self.p + me]
                self.ci = [0, wilson_ci(stats.norm.ppf(self.conf))[1]]
            else:
                self.p_val = 1 - p_val
                # traditional CI calculation
                # z_critical = stats.norm.ppf(self.conf)
                # me = z_critical * self.se
                # self.ci = [self.p - me, 1]
                self.ci = [wilson_ci(stats.norm.ppf(self.conf))[0], 1]

            # can't replicate R's z.value with statsmodels
            # var_p0 = self.comp_value * (1 - self.comp_value)
            # self.z_score, self.p_val = proportion.proportions_ztest(
            #     self.ns, self.n, self.comp_value, prop_var=var_p0
            # )

    def summary(self, dec=3) -> None:
        print(
            f'Single proportion ({ifelse(self.test_type=="binomial", "binomial exact", "z-test")})'
        )
        print(f"Data      : {self.name}")
        print(f"Variable  : {self.var}")
        print(f'Level     : "{self.lev}" in {self.var}')
        print(f"Confidence: {self.conf}")
        print(
            f'Null hyp. : the proportion of "{self.lev}" in {self.var} is equal to {self.comp_value}'
        )

        if self.alt_hyp == "less":
            alt_hyp = "less than"
        elif self.alt_hyp == "two-sided":
            alt_hyp = "not equal to"
        else:
            alt_hyp = "greater than"

        cl = bu.ci_label(self.alt_hyp, self.conf, dec=dec)

        print(
            f'Alt. hyp. : the proportion of "{self.lev}" in {self.var} {alt_hyp} {self.comp_value}\n'
        )

        stats_df = pd.DataFrame(
            {
                "p": [self.p],
                "ns": [self.ns],
                "n": [self.n],
                "n_missing": [self.n_missing],
                "sd": [self.sd],
                "se": [self.se],
                "me": [self.me],
            }
        ).round(dec)

        statistic = [
            ifelse(self.test_type == "binomial", "ns", "z.value"),
            ifelse(self.test_type == "binomial", [self.ns], [self.z_score]),
        ]
        test_df = pd.DataFrame(
            {
                "diff": [self.diff],
                statistic[0]: statistic[1],
                "p.value": [ifelse(self.p_val < 0.001, "< .001", self.p_val)],
                cl[0]: [self.ci[0]],
                cl[1]: [self.ci[1]],
                "": [sig_stars([self.p_val])[0]],
            }
        ).round(dec)

        print(stats_df.to_string(index=False), "\n")
        print(test_df.to_string(index=False))
        print("\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1")

    def plot(self, plots: str = "bar") -> None:
        if plots == "bar":
            proportion_data = self.data[self.var].value_counts(normalize=True)
            ax = proportion_data.plot(kind="bar", alpha=0.5)
            ax.set_ylabel("")
            ax.set_title(f'Single proportion: "{self.lev}" in {self.var}')
            ax.yaxis.set_major_formatter(lambda x, _: "{:.0%}".format(x))


if __name__ == "__main__":
    import pyrsm as rsm

    consider, consider_description = rsm.load_data(pkg="basics", name="consider")
    sp = single_prop(
        data={"consider": consider},
        var="consider",
        lev="yes",
        alt_hyp="less",
        conf=0.95,
        comp_value=0.1,
        test_type="binomial",
    )
    sp.summary()
    sp.plot()



from typing import Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import polars as pl
import seaborn as sns
from scipy import stats

from pyrsm.model.model import sig_stars
from pyrsm.utils import check_dataframe


class correlation:
    """
    Calculate correlations between numeric variables in a Pandas dataframe

    Parameters
    ----------
    data : Pandas dataframe with numeric variables

    Returns
    -------
    Correlation object with two key attributes
    cr: Correlation matrix
    cp: p.value matrix
    cv: Covariance matrix

    Examples
    --------
    import pandas as pd
    import pyrsm as rsm
    salary, salary_description = rsm.load_data(pkg="basics", name="salary")
    cr = rsm.correlation(salary[["salary", "yrs.since.phd", "yrs.service"]])
    cr.cr
    """

    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        vars: Optional[list[str]] = [],
        method: str = "pearson",
    ) -> None:
        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name].copy()
        else:
            self.data = data
            self.name = "Not provided"

        self.data = check_dataframe(self.data)
        self.vars = vars
        if len(self.vars) == 0:
            self.vars = [
                col
                for col in self.data.columns
                if pd.api.types.is_numeric_dtype(self.data[col].dtype)
            ]
        self.data = self.data[self.vars].copy()
        self.method = method

        ncol = self.data.shape[1]
        cr = np.zeros([ncol, ncol])
        cp = cr.copy()
        cv = cr.copy()
        for i in range(ncol - 1):
            for j in range(i + 1, ncol):
                cdf = self.data.iloc[:, [i, j]]
                # pairwise deletion
                cdf = cdf[~np.any(np.isnan(cdf), axis=1)]
                # convert bool to int to avoid errors in scipy
                cdf = cdf.astype(
                    {
                        col: int
                        for col in cdf.columns
                        if cdf[col].dtype == bool
                    }
                )
                if self.method == "spearman":
                    c = stats.spearmanr(cdf.iloc[:, 0], cdf.iloc[:, 1])
                elif self.method == "kendall":
                    c = stats.kendalltau(cdf.iloc[:, 0], cdf.iloc[:, 1])
                else:
                    c = stats.pearsonr(cdf.iloc[:, 0], cdf.iloc[:, 1])

                cr[j, i] = c[0]
                cp[j, i] = c[1]
                cv[j, i] = cdf.iloc[:, [0, 1]].cov().iloc[0, 1]

        self.cr = cr
        self.cp = cp
        self.cv = cv

    def summary(self, cov=False, cutoff: float = 0, dec: int = 2) -> None:
        """
        Print correlations between numeric variables in a Pandas dataframe

        Parameters
        ----------
        cov : bool
            Show the covariance matrix if set to True
        cutoff : float
            Only show correlations larger than a threshold in absolute value
        dec : int
            Number of decimal places to use in rounding

        Examples
        --------
        import pandas as pd
        import pyrsm as rsm
        salary, salary_description = load_data(pkg="basics", name="salary")
        cr = rsm.correlation(salary[["salary", "yrs.since.phd", "yrs.service"]])
        cr.summary()
        """
        prn = "Correlation\n"
        prn += f"Data     : {self.name}\n"
        prn += f"Method   : {self.method}"
        print(prn)
        if len(self.vars) < 2:
            print("\n**Select two or more variables to calculate correlations**")
        else:
            ind = np.triu_indices(self.cr.shape[0])
            cn = list(self.data.columns)

            # correlations
            crs = pd.DataFrame(self.cr.round(dec).astype(str), columns=cn, index=cn)
            if cutoff > 0:
                crs.values[np.abs(self.cr) < cutoff] = ""
            crs.values[ind] = ""

            # p.values
            cps = pd.DataFrame(self.cp.round(dec).astype(str), columns=cn, index=cn)
            if cutoff > 0:
                cps.values[np.abs(self.cr) < cutoff] = ""
            cps.values[ind] = ""

            if len(cn) > 2:
                x = "x"
                y = "y"
            else:
                x = cn[0]
                y = cn[1]

            prn = f"Cutoff   : {cutoff}\n"
            prn += "Variables: " + ", ".join(cn) + "\n"
            prn += f"Null hyp.: variables {x} and {y} are not correlated\n"
            prn += f"Alt. hyp.: variables {x} and {y} are correlated\n"
            print(prn)
            print("Correlation matrix:")
            print(crs.iloc[1:, :-1])
            print("\np.values:")
            print(cps.iloc[1:, :-1])

            if cov:
                cvs = pd.DataFrame(self.cv.round(dec), columns=cn, index=cn).map(
                    lambda x: "{:,}".format(x)
                )
                if cutoff > 0:
                    cvs.values[np.abs(self.cr) < cutoff] = ""

                cvs.values[ind[0], ind[1]] = ""
                print("\nCovariance matrix:")
                print(cvs.iloc[1:, :-1])

    def plot(self, nobs: int = 1000, dec: int = 2, figsize: tuple[float, float] = None):
        """
        Plot of correlations between numeric variables in a Pandas dataframe

        Parameters
        ----------
        nobs : int
            Number of observations to use for the scatter plots. The default
            value is 1,000. To use all observations in the plots, use nobs=-1
        dec : int
            Number of decimal places to use in rounding
        figsize : tuple
            A tuple that determines the figure size. If None, size is
            determined based on the number of numeric variables in the
            data

        Examples
        --------
        import pandas as pd
        import pyrsm as rsm
        rsm.(pkg="basics", name="salary", dct=globals())
        cr = rsm.correlation(salary[["salary", "yrs.since.phd", "yrs.service"]])
        cr.plot(figsize=(7, 3))
        """

        if figsize is None:
            figsize = (max(5, self.cr.shape[0]), max(self.cr.shape[0], 5))

        def cor_label(label, longest, ax_sub):
            ax_sub.axes.xaxis.set_visible(False)
            ax_sub.axes.yaxis.set_visible(False)
            # set font size to avoid exceeding boundaries
            fs = min(figsize[0], figsize[1])
            font = (80 * fs) / (len(longest) * self.cr.shape[0])
            ax_sub.text(
                0.5,
                0.5,
                label,
                horizontalalignment="center",
                verticalalignment="center",
                fontsize=font,
            )

        def cor_text(r, p, ax_sub, dec=2):
            if np.isnan(p):
                p = 1

            p = round(p, dec)
            rt = round(r, dec)
            p_star = sig_stars([p])[0]

            fs = min(figsize[0], figsize[1])
            font = (50 * fs) / (len(str(rt)) * len(self.data.columns))
            font_star = (12 * fs) / len(self.data.columns)

            ax_sub.axes.xaxis.set_visible(False)
            ax_sub.axes.yaxis.set_visible(False)
            ax_sub.text(
                0.5,
                0.5,
                rt,
                horizontalalignment="center",
                verticalalignment="center",
                fontsize=font * abs(r),
            )
            ax_sub.text(
                0.8,
                0.8,
                p_star,
                horizontalalignment="center",
                verticalalignment="center",
                fontsize=font_star,
                color="blue",
            )

        def cor_plot(x_data, y_data, ax_sub, s_size):
            sns.regplot(
                x=x_data,
                y=y_data,
                ax=ax_sub,
                scatter_kws={"alpha": 0.3, "color": "black", "s": s_size},
                line_kws={"color": "blue"},
            )

            ax_sub.axes.xaxis.set_visible(False)
            ax_sub.axes.yaxis.set_visible(False)

        def cor_mat(data, cmat, pmat, dec=2, nobs=1000, figsize=None):
            cn = data.columns
            ncol = len(cn)
            longest = max(cn, key=len)
            s_size = 50 / len(self.data.columns)

            fs = min(figsize[0], figsize[1])
            s_size = (5 * fs) / len(self.data.columns)

            _, axes = plt.subplots(ncol, ncol, figsize=figsize)

            if nobs < data.shape[0] and nobs != np.inf and nobs != -1:
                data = data.copy().sample(nobs)

            for i in range(ncol):
                for j in range(ncol):
                    if i == j:
                        cor_label(cn[i], longest, axes[i, j])
                    elif i > j:
                        cor_plot(data[cn[i]], data[cn[j]], axes[i, j], s_size)
                    else:
                        cor_text(cmat[j, i], pmat[j, i], axes[i, j], dec=2)

            plt.subplots_adjust(wspace=0.04, hspace=0.04)

        cor_mat(self.data, self.cr, self.cp, dec=dec, nobs=nobs, figsize=figsize)



import io
from contextlib import redirect_stderr, redirect_stdout

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from shiny import reactive, render, req, ui

import pyrsm as rsm
import pyrsm.radiant.utils as ru
from pyrsm.utils import intersect


def ui_predict(self, show_ci=True):
    if show_ci:
        ci = (
            ui.input_checkbox(
                "pred_ci",
                "Show conf. intervals",
                value=self.state.get("pred_ci", False),
            ),
            ui.panel_conditional(
                "input.pred_ci == true",
                ui.input_slider(
                    id="conf",
                    label="Confidence level:",
                    min=0,
                    max=1,
                    value=self.state.get("conf", 0.95),
                ),
            ),
        )
    else:
        ci = None

    return ui.panel_conditional(
        "input.tabs == 'Predict'",
        ui.panel_well(
            ui.input_select(
                "pred_type",
                "Prediction input type:",
                ["None", "Data", "Command", "Data & Command"],
                selected=self.state.get("pred_type", "None"),
            ),
            ui.panel_conditional(
                "input.pred_type == 'Data' || input.pred_type == 'Data & Command'",
                ui.input_select(
                    "pred_datasets",
                    "Prediction data:",
                    self.dataset_list,
                    selected=self.state.get("pred_data", None),
                ),
            ),
            ui.panel_conditional(
                "input.pred_type == 'Command' || input.pred_type == 'Data & Command'",
                ru.input_return_text_area(
                    "pred_cmd",
                    "Prediction command:",
                    rows=3,
                    value=self.state.get("pred_cmd", ""),
                    placeholder="Specify a dictionary of values to to use for prediction, e.g., {'carat': 1, 'cut': 'Ideal'}",  # noqa: E501
                ),
            ),
            ci,
        ),
    )


def make_model_inputs(self, input, output, get_data, mod_type="isBin"):
    if isinstance(mod_type, str):

        @output(id="ui_rvar")
        @render.ui
        def ui_rvar():
            isType = get_data()["var_types"][mod_type]
            return ui.input_select(
                id="rvar",
                label="Response Variable:",
                selected=self.state.get("rvar", None),
                choices=isType,
            )

    elif isinstance(mod_type, dict):

        @output(id="ui_rvar")
        @render.ui
        def ui_rvar():
            isType = get_data()["var_types"][mod_type[input.mod_type()]]
            return ui.input_select(
                id="rvar",
                label="Response Variable:",
                selected=self.state.get("rvar", None),
                choices=isType,
            )

    @output(id="ui_evar")
    @render.ui
    def ui_evar():
        vars = get_data()["var_types"]["all"].copy()
        if (input.rvar() is not None) and (input.rvar() in vars):
            del vars[input.rvar()]

        return ui.input_select(
            id="evar",
            label="Explanatory Variables:",
            selected=self.state.get("evar", []),
            choices=vars,
            multiple=True,
            size=min(8, len(vars)),
            selectize=False,
        )


def make_int_inputs(self, input, output, get_data):
    @output(id="ui_interactions")
    @render.ui
    def ui_interactions():
        if len(input.evar()) > 1 and input.show_interactions() is not None:
            choices = []
            nway = int(input.show_interactions())
            isNum = intersect(
                list(input.evar()), list(get_data()["var_types"]["isNum"].keys())
            )
            if len(isNum) > 0:
                # choices += ru.qterms(isNum, nway=int(input.show_interactions()[0]))
                choices += ru.qterms(isNum, nway=nway)
            choices += ru.iterms(input.evar(), nway=nway)
            return ui.input_select(
                id="interactions",
                label=None,
                selected=self.state.get("interactions", None),
                choices=choices,
                multiple=True,
                size=min(8, len(choices)),
                selectize=False,
            )

    @output(id="ui_evar_test")
    @render.ui
    def ui_evar_test():
        choices = {e: e for e in input.evar()}
        if (
            "show_interactions" in input
            and input.show_interactions() is not None
            and int(input.show_interactions()[0]) > 1
        ):
            choices.update({e: e for e in input.interactions()})
        return ui.input_selectize(
            id="evar_test",
            label="Variables to test:",
            selected=self.state.get("evar_test", None),
            choices=choices,
            multiple=True,
        )

    @output(id="ui_incl_evar")
    @render.ui
    def ui_incl_evar():
        if len(input.evar()) > 0:
            return ui.input_select(
                id="incl_evar",
                label="Explanatory variables to include:",
                selected=self.state.get("incl_evar", None),
                choices=input.evar(),
                multiple=True,
                size=min(8, len(input.evar())),
                selectize=False,
            )

    @output(id="ui_incl_interactions")
    @render.ui
    def ui_incl_interactions():
        if len(input.evar()) > 1:
            if "show_interactions" in input and input.show_interactions() is not None:
                nway = int(input.show_interactions())
            else:
                nway = 2
            choices = ru.iterms(input.evar(), nway=nway)
            return ui.input_select(
                id="incl_interactions",
                label="Interactions to include:",
                selected=self.state.get("incl_interactions", None),
                choices=choices,
                multiple=True,
                size=min(8, len(choices)),
                selectize=False,
            )


def make_estimate(
    self,
    input,
    output,
    get_data,
    fun,
    ret,
    module="",
    run=True,
    ec=None,
    debug=True,
):
    if ec is None:

        def estimation_code():
            data_name, code = (get_data()[k] for k in ["data_name", "code"])
            # from https://discord.com/channels/1109483223987277844/1127817202804985917/1129530385429176371
            # does not work with ifelse for some reason
            # inp = ifelse("lev" in input, input.lev(), None)
            if "lev" in input:
                inp = input.lev()
            else:
                inp = None
            if "weights" in input:
                weights = input.weights()
                if weights == "None":
                    weights = None
            else:
                weights = None

            args = {
                "data": f"""{{"{data_name}": {data_name}}}""",
                "rvar": input.rvar(),
                "lev": inp,
                "evar": list(input.evar()),
                "weights": weights,
            }

            if (
                "show_interactions" in input
                and input.show_interactions() is not None
                and int(input.show_interactions()) > 0
                and len(input.interactions()) > 0
            ):
                args["ivar"] = list(input.interactions())

            if ret == "xgb":
                args.update(
                    {
                        "mod_type": input.mod_type(),
                        "n_estimators": input.n_estimators(),
                        "max_depth": input.max_depth(),
                        "min_child_weight": input.min_child_weight(),
                        "learning_rate": input.learning_rate(),
                        "subsample": input.subsample(),
                        "colsample_bytree": input.colsample_bytree(),
                        "random_state": input.random_state(),
                    }
                )
            elif ret == "rf":
                if input.max_features() in ["sqrt", "log2"]:
                    max_features = input.max_features()
                else:
                    max_features = int(input.max_features())

                args.update(
                    {
                        "mod_type": input.mod_type(),
                        "n_estimators": input.n_estimators(),
                        "max_features": max_features,
                        "min_samples_leaf": input.min_samples_leaf(),
                        "max_samples": input.max_samples(),
                        "random_state": input.random_state(),
                    }
                )
            elif ret == "nn":
                args.update(
                    {
                        "mod_type": input.mod_type(),
                        "hidden_layer_sizes": eval(input.hidden_layer_sizes()),
                        "activation": input.activation(),
                        "solver": input.solver(),
                        "alpha": input.alpha(),
                        # "batch_size": input.batch_size(),
                        "learning_rate_init": input.learning_rate_init(),
                        "max_iter": input.max_iter(),
                    }
                )

            args_str = ru.drop_default_args(args, getattr(eval(f"rsm.{module}"), fun))

            if "extra_args" in input and input.extra_args() != "":
                args_str += ", " + input.extra_args()
            return f"""rsm.{module}.{fun}({args_str})""", code

    else:
        estimation_code = ec

    if debug:

        @output(id="show_estimation_code")
        @render.text
        def show_estimation_code():
            out = io.StringIO()
            with redirect_stdout(out), redirect_stderr(out):
                eval(estimation_code()[0])

            return out.getvalue()

    def show_code():
        sc = estimation_code()
        return f"""{sc[1]}\n{ret} = {sc[0]}"""

    if run:

        @reactive.Calc
        @reactive.event(input.run, ignore_none=True)
        def estimate():
            locals()[input.datasets()] = self.datasets[
                input.datasets()
            ]  # get data into local scope

            ec = estimation_code()
            # exec for multi-line
            # https://stackoverflow.com/a/12698067/1974918
            exec(ec[1])
            return eval(ec[0])

    else:

        @reactive.Calc
        def estimate():
            locals()[input.datasets()] = self.datasets[
                input.datasets()
            ]  # get data into local scope

            ec = estimation_code()
            exec(ec[1])
            return eval(ec[0])

    return show_code, estimate


def make_summary(
    self, input, output, session, show_code, estimate, ret, sum_fun, sc=None
):
    if sc is None:

        def summary_code():
            if "controls" in input:
                args = {c: True for c in input.controls()}
            else:
                args = {}
            if (
                "evar_test" in input
                and input.evar_test() is not None
                and len(input.evar_test()) > 0
            ):
                args["test"] = list(input.evar_test())

            args_string = ru.drop_default_args(args, sum_fun)
            return f"""{ret}.summary({args_string})"""

    else:
        summary_code = sc

    @output(id="show_summary_code")
    @render.text
    def show_summary_code():
        cmd = f"""{show_code()}\n{summary_code()}"""
        return ru.code_formatter(cmd, self, input, session, id="copy_summary")

    @output(id="summary")
    @render.text
    def summary():
        out = io.StringIO()
        with redirect_stdout(out), redirect_stderr(out):
            locals()[ret] = estimate()  # get model object into local scope
            eval(summary_code())
        return out.getvalue()


def make_predict(
    self, input, output, session, show_code, estimate, ret, pred_fun, show_ci=False
):
    def predict_code():
        args = {}
        cmd = input.pred_cmd().strip()

        if input.pred_type() == "Data":
            args["data"] = input.pred_datasets()
        elif input.pred_type() == "Command" and not ru.is_empty(cmd):
            args["data"] = None
            args["cmd"] = cmd
        elif input.pred_type() == "Data & Command" and not ru.is_empty(cmd):
            args["data"] = input.pred_datasets()
            args["data_cmd"] = cmd

        if "pred_ci" in input:
            ci = input.pred_ci()
            if ci:
                args.update({"ci": ci, "conf": input.conf()})

        args_string = ru.drop_default_args(
            args, pred_fun, ignore=["data", "cmd", "data_cmd"]
        )
        return f"""{ret}.predict({args_string})"""

    @output(id="show_predict_code")
    @render.text
    def show_predict_code():
        return ru.code_formatter(
            f"""{show_code()}\npred = {predict_code()}""",
            self,
            input,
            session,
            id="copy_predict",
        )

    @output(id="predict")
    @render.data_frame
    def predict():
        if input.pred_type() != "None":
            locals()[ret] = estimate()  # get model object into local scope
            if input.pred_type() in ["Data", "Data & Command"]:
                locals()[input.pred_datasets()] = self.datasets[
                    input.pred_datasets()
                ]  # get prediction data into local scope
            pred = eval(predict_code())

            summary = "Viewing rows {start} through {end} of {total}"
            if pred.shape[0] > 100_000:
                pred = pred[:100_000]
                summary += " (100K rows shown)"

            return render.DataTable(pred.round(3), summary=summary)
        else:
            return None


def make_plot(self, input, output, session, show_code, estimate, ret, pc=None):
    if pc is None:

        def plot_code():
            plots = input.plots()
            if plots == "pred":
                incl = list(input.incl_evar())
                if "incl_interactions" in input:
                    incl_int = list(input.incl_interactions())
                else:
                    incl_int = []
                cmd = f""", incl={incl}"""
                if len(incl_int) > 0:
                    cmd += f""", incl_int={incl_int}"""
                cmd = f"""{ret}.plot(plots="{plots}" {cmd})"""
            elif plots == "corr":
                cmd = f"""{ret}.plot(plots="{plots}", nobs={input.nobs()})"""
            else:
                cmd = f"""{ret}.plot(plots="{plots}")"""
            return cmd

    else:
        plot_code = pc

    @output(id="show_plot_code")
    @render.text
    def show_plot_code():
        plots = input.plots()
        if plots != "None":
            cmd = f"""{show_code()}\n{plot_code()}"""
            return ru.code_formatter(cmd, self, input, session, id="copy_plot")

    @reactive.Calc
    def gen_plot():
        locals()[ret] = estimate()
        eval(f"""{plot_code()}""")
        fig = plt.gcf()
        width, height = fig.get_size_inches()  # Get the size in inches
        return fig, width * 96, height * 96

    @output(id="plot")
    @render.plot
    def plot():
        plots = input.plots()
        if plots != "None":
            return gen_plot()[0]

    @output(id="plot_container")
    @render.ui
    def plot_container():
        req(estimate(), input.plots())
        width, height = gen_plot()[1:]
        aspect_ratio = height / width
        # return ui.output_plot("plot", height=f"{height}px", width=f"{width}px")
        return ui.div(
            {"style": f"width: 100%; padding-bottom: {aspect_ratio * 100}%; position: relative;"},
            ui.div(
                {"style": "position: absolute; top: 0; left: 0; right: 0; bottom: 0;"},
                ui.output_plot("plot", height="100%", width="100%"),
            ),
        )
